import os
import logging

import pytest
import IPython

from hxnfly import ppmac_util

from .fly_mock import (MockIPython, MockPpmac, MockPositioner, MockLogbook,
                       AttrDict)
from .sim_detector import TestDetector


LOG_SETUP = False


@pytest.fixture(scope='function')
def ipython(monkeypatch):
    global LOG_SETUP

    def mock_ipython():
        return _ipython

    _ipython = MockIPython()
    monkeypatch.setattr(IPython, 'get_ipython', mock_ipython)

    import hxnfly.log

    if not LOG_SETUP:
        log_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                'hxnfly.log')
        hxnfly.log.log_setup(log_path=log_path)
        LOG_SETUP = True

    _ipython.user_ns['logbook'] = MockLogbook()
    return _ipython


@pytest.fixture(scope='function')
def logbook(ipython):
    return ipython.user_ns['logbook']


@pytest.fixture(scope='function', autouse=True)
def ppmac(monkeypatch, ipython):
    def mock_connect(*args, **kwargs):
        return _ppmac

    _ppmac = MockPpmac()
    monkeypatch.setattr(ppmac_util, 'ppmac_connect', mock_connect)
    import hxnfly.fly
    monkeypatch.setattr(hxnfly.fly, 'ppmac_connect', mock_connect)
    return _ppmac


@pytest.fixture(scope='function')
def gpascii(ppmac):
    return ppmac.gpascii


def set_motor_status(gpascii, axis, home_pos, act_pos, in_position,
                     closed_loop):
    gpascii.set_variable('motor[{}].homepos'.format(axis), float(home_pos))
    gpascii.set_variable('motor[{}].actpos'.format(axis), float(act_pos))
    gpascii.set_variable('motor[{}].inpos'.format(axis), int(in_position))
    gpascii.set_variable('motor[{}].closedloop'.format(axis), int(closed_loop))


@pytest.fixture(scope='function')
def axes(gpascii):
    axes = dict(testx=30, testy=31, testz=32)
    for num in axes.values():
        set_motor_status(gpascii, num, home_pos=0.0, act_pos=1.0,
                         in_position=True, closed_loop=True)

    gpascii.positioners = {name: MockPositioner(name=name)
                           for name, num in axes.items()}
    gpascii.positioners_by_number = {num: gpascii.positioners[name]
                                     for name, num in axes.items()}
    return axes


@pytest.fixture(scope='function')
def positioners(gpascii, axes):
    return gpascii.positioners


@pytest.fixture(scope='module')
def sim_det():
    det = TestDetector('', name='test_det', image_name='sim_tiff')
    det.wait_for_connection()
    return det


@pytest.fixture(scope='function')
def xspress3(ipython):
    return ipython.user_ns['xspress3']


@pytest.fixture(scope='function')
def run_engine(ipython, monkeypatch):
    from bluesky.tests.utils import setup_test_run_engine

    run_engine = setup_test_run_engine()

    new_md = dict(run_engine.md)
    new_md.update(ipython.user_ns['gs'].RE.md)
    run_engine.md = new_md

    run_engine.ignore_callback_exceptions = False
    return run_engine


@pytest.fixture(scope='function')
def global_state(monkeypatch, run_engine):
    import hxntools.scans

    global_state = AttrDict(RE=run_engine, detectors=[])
    monkeypatch.setattr(hxntools.scans, 'get_gs', lambda: global_state)

    import hxntools.scans
    hxntools.scans.setup(debug_mode=True)
    return global_state


def setup_function(function):
    logger = logging.getLogger('hxnfly')
    logger.info('')
    logger.info('----------------------------------')
    logger.info('{}'.format(function))


def teardown_function(function):
    logger = logging.getLogger('hxnfly')
    logger.info('----------------------------------')
    logger.info('')


def setup_module(module):
    for name in ['hxnfly', 'hxntools', 'bluesky']:
        logging.getLogger(name).setLevel(logging.DEBUG)

    import filestore
    import filestore.handlers
    filestore.api.register_handler('AD_TIFF',
                                   filestore.handlers.AreaDetectorTiffHandler)



import os

from ophyd.areadetector.filestore_mixins import (FileStoreIterativeWrite,
                                                 FileStoreTIFF)

from ophyd import (Device, Component as Cpt)
from ophyd.areadetector import (SimDetectorCam, HDF5Plugin, TIFFPlugin,
                                DetectorBase)
from hxntools.detectors.trigger_mixins import (HxnModalTrigger,
                                               FileStoreBulkReadable)


DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                         'test_data')


class SimTIFFPlugin(TIFFPlugin, FileStoreBulkReadable, FileStoreTIFF, Device):
    def __init__(self, prefix, **kwargs):
        print('tiff plugin', prefix, kwargs)
        super().__init__(prefix, **kwargs)

    def mode_external(self):
        print('tiff mode external')
        total_points = self.parent.mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

    def get_frames_per_point(self):
        mode = self.parent.mode_settings.mode.get()
        if mode == 'external':
            return 1
        else:
            return self.parent.cam.num_images.get()


class TestDetector(HxnModalTrigger, DetectorBase):
    tiff1 = Cpt(SimTIFFPlugin, 'XF:31IDA-BI{Cam:Tbl}TIFF1:',
                read_attrs=[],
                configuration_attrs=[],
                write_path_template=os.path.join(DATA_PATH, '%Y-%m-%d', ''),
                )

    cam = Cpt(SimDetectorCam, 'XF:31IDA-BI{Cam:Tbl}cam1:')
    # hdf5 = Cpt(SimDetectorCam, 'XF:31IDA-BI{Cam:Tbl}HDF5:')

    def mode_external(self):
        self.tiff1.stage_sigs[self.tiff1.blocking_callbacks] = 1
        self.cam.stage_sigs[self.cam.acquire_time] = 0.01
        self.cam.stage_sigs[self.cam.acquire_period] = 0.10
        super().mode_external()



import re
import time
import logging

import numpy as np

import ophyd
from ophyd import (Signal, Component as Cpt)
from hxntools.detectors.trigger_mixins import HxnModalBase


class AttrDict(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__dict__ = self


class MockFastGather:
    def __init__(self, ppmac):
        self.ppmac = ppmac

        # state for the test:
        self.gathered = None
        self.n_addr = None
        self.n_samples = None

    def query_types(self):
        return 'type_placeholder'

    def query_raw_data(self):
        return 1, b'abcd'

    def _parse_raw_data(self, dtypes, raw_data):
        return self.gathered, self.n_addr, self.n_samples

    def __repr__(self):
        return 'MockFastGather()'


class MockGpascii:
    def __init__(self, ppmac):
        self.ppmac = ppmac
        self.servo_period = 1.0

        # for holdint test state:
        self.variables = {}
        self.positioners = {}
        self.positioners_by_number = {}

    def run_and_wait(self, coord_sys, prog_num, verbose=False, **kwargs):
        # NOTE: need enough time for areadetector to write out the frames:
        time.sleep(1.0)
        return 0  # success!

    def program(self, coord_sys, prog_num, stop=False, **kwargs):
        print('program kwargs', kwargs)

    def send_program(self, coord_sys, prog_num, filename=None, verbose=None,
                     macros=None, motors=None):
        print('send_program', coord_sys, prog_num, filename, verbose, macros,
              motors)

    def run_simple_script(self, script_fn):
        print('run_simple_script', script_fn)

    def get_variable(self, name, type_=float):
        return type_(self.variables[name.lower()])

    def set_variable(self, name, value):
        self.variables[name.lower()] = value

    def send_line(self, line):
        m = re.match('#(\d+)j=(.*)', line)
        if m:
            motor, position = m.groups()
            motor = int(motor)
            position = float(position)
            self.jog(motor, position, relative=False)

    def jog(self, motor, target, relative=False, wait=False):
        if relative:
            target += self.positioners_by_number[motor].position

        self.positioners_by_number[motor].position = target
        print('moved motor', motor, 'to position', target)

    def sync(self):
        pass

    def __repr__(self):
        return 'MockGpascii()'


class MockPpmac:
    def __init__(self, *args, **kwargs):
        self.fast_gather = MockFastGather(self)
        self.gpascii = MockGpascii(self)

    def __repr__(self):
        return 'MockPpmac()'


class MockDetector(HxnModalBase):
    def stage(self):
        print('staging', self, self.mode_settings.get())

    def unstage(self):
        print('unstaging', self)


class MockMca(ophyd.Device):
    spectrum = Cpt(Signal, value=np.array([1, 2, 3]))

    def __init__(self, ch, name):
        super().__init__('', name=name)
        self.ch = ch
        self.spectrum.pvname = 'fake_mca_pvname_{}'.format(ch)


class MockScaler(HxnModalBase):
    mca_by_index = {ch: MockMca(ch, name='mca{}'.format(ch))
                    for ch in range(32)}

    def stage(self):
        print('staging', self, self.mode_settings.get())

    def unstage(self):
        print('unstaging', self)


class MockPositioner:
    def __init__(self, name=''):
        self.name = name
        self.position = 0.0
        self.egu = 'um'

    def move(self, position, wait=False):
        raise ValueError()

    def __repr__(self):
        return 'MockPositioner(name={!r})'.format(self.name)


class MockSignal:
    def __init__(self, name, value=None):
        self.name = name
        self.value = value
        self._timestamp = 1

    def get(self, *, count=None, use_monitor=False):
        if count is not None:
            return self.value[:count]

        return self.value

    def put(self, value):
        self.value = value

    @property
    def timestamp(self):
        self._timestamp += 1
        return self._timestamp


class MockXspress3ROI:
    def __init__(self, name, value):
        self.name = name
        self.value = MockSignal(name, [1, 2, 3])
        self.value.put(value)

    channel_num = 1


class MockXspress3HDF5:
    _fn = '_fake_staged_filename_'
    capture = MockSignal('capture', 0)


class MockXspress3(MockDetector):
    def __init__(self, name):
        super().__init__(name, name=name)

    def setup_fake_rois(self, rois):
        self.enabled_rois = [MockXspress3ROI(name, spectrum)
                             for name, spectrum in rois]

    def read_hdf5(self, fn):
        for roi in self.enabled_rois:
            # yield roi name, namedtuple with info + spectrum
            yield roi.name, AttrDict(value=roi.value.get(),
                                     ev_low=5, ev_high=10,
                                     )

    enabled_rois = [MockXspress3ROI('Fe', [1, 2, 3]),
                    MockXspress3ROI('Mo', [3, 2, 1]),
                    ]
    acquire = MockSignal('acquire', 0)
    settings = AttrDict(array_counter=MockSignal('array_counter', 3))
    external_trig = MockSignal('external_trig', True)
    hdf5 = MockXspress3HDF5()
    spectra_per_point = MockSignal('spectra_per_point', 1)
    total_points = MockSignal('total_points', 3)


class MockIPython:
    # TODO: scan_id may be off by 1 with RE sync?
    user_ns = {'gs': AttrDict(RE=AttrDict(md={'scan_id': 99999})),
               'zebra': MockDetector('', name='zebra'),
               'sclr1': MockScaler('', name='sclr1'),
               'ssx': MockPositioner(name='ssx'),
               'ssy': MockPositioner(name='ssy'),
               'ssz': MockPositioner(name='ssz'),
               'zpssx': MockPositioner(name='zpssx'),
               'zpssy': MockPositioner(name='zpssy'),
               'zpssz': MockPositioner(name='zpssz'),
               'xspress3': MockXspress3(name='xspress3'),
               }

    class MockEvents:
        def register(self, event, callback):
            pass

    events = MockEvents()


class MockLogbook:
    def log(self, log_text, **kwargs):
        logger = logging.getLogger('hxnfly.fly')
        logger.info('[[LOGBOOK]] {}'.format(log_text))
